{% extends "base.html" %}

{% block navigation %}
    <ts-sketch-navigation sketch-id="{{ sketch.id }}" active="explore" graphs-enabled="{{ graphs_enabled }}"></ts-sketch-navigation>
{% endblock %}

{% block main %}
        {% if sketch.has_permission(user=current_user, permission='write') and not sketch.timelines %}
            <div class="row">
                <div class="col-md-12">
                    <div class="card">
                        <div class="text-center" style="padding:100px;">
                            <h3 style="color:#777"><i class="fa fa-meh-o"></i> No data to analyze</h3>
                            <a class="btn btn-success" href="{{ url_for('sketch_views.timelines', sketch_id=sketch.id) }}"><i class="fa fa-plus"></i> Add timeline</a>
                        </div>
                    </div>
                </div>
            </div>
        {% else %}
            <div class="row">
                <div class="col-md-12">
                    <ts-search sketch-id="{{ sketch.id }}" view-id="{{ view.id }}" named-view="{{ named_view }}" searchtemplate-id="{{ searchtemplate_id }}" similarity-enabled="{{ similarity_enabled }}" autoload="true" redirect="true"></ts-search>
                </div>
            </div>
        {% endif %}
{% endblock %}
